@@ -3,7 +3,7 @@
  * Ruby Cairo Binding
  *
  * $Author: kou $
- * $Date: 2005/10/10 15:40:26 $
+ * $Date: 2006/05/02 05:37:02 $
  *
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -46,10 +46,18 @@ rb_cairo_font_face_to_ruby_object (cairo_font_face_t *face)
     }
 }
 
+static VALUE
+cr_font_face_get_type (VALUE self)
+{
+  return INT2NUM ( cairo_font_face_get_type (_SELF));
+}
+
 
 void
 Init_cairo_font (void)
 {
   rb_cCairo_FontFace =
     rb_define_class_under (rb_mCairo, "FontFace", rb_cObject);
+
+  rb_define_method (rb_cCairo_FontFace, "type", cr_font_face_get_type, 0);
 }
