@@ -3,7 +3,7 @@
  * Ruby Cairo Binding
  *
  * $Author: kou $
- * $Date: 2005/10/15 14:33:46 $
+ * $Date: 2006/07/01 14:45:55 $
  *
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -15,12 +15,10 @@
 #include "rb_cairo.h"
 #include "rubyio.h"
 
-#if CAIRO_HAS_PS_SURFACE
-#include <cairo-ps.h>
-#endif
-
-#if CAIRO_HAS_PDF_SURFACE
-#include <cairo-pdf.h>
+#if CAIRO_HAS_PS_SURFACE || CAIRO_HAS_PDF_SURFACE || CAIRO_HAS_SVG_SURFACE
+#  define HAS_CREATE_CR_CLOSURE_SURFACE 1
+#else
+#  define HAS_CREATE_CR_CLOSURE_SURFACE 0
 #endif
 
 
@@ -28,6 +26,7 @@ VALUE rb_cCairo_Surface;
 VALUE rb_cCairo_ImageSurface;
 VALUE rb_cCairo_PDFSurface;
 VALUE rb_cCairo_PSSurface;
+VALUE rb_cCairo_SVGSurface;
 
 static ID cr_id_read;
 static ID cr_id_write;
@@ -60,11 +59,33 @@ cr_surface_set_klass (cairo_surface_t *surface, VALUE klass)
 static VALUE
 cr_surface_get_klass (cairo_surface_t *surface)
 {
-  VALUE klass = (VALUE)cairo_surface_get_user_data (surface, &cr_klass_key);
-  if (!klass)
+  VALUE klass;
+  void *data = cairo_surface_get_user_data (surface, &cr_klass_key);
+
+  if (data)
+    klass = (VALUE) data;
+  else
     {
-      rb_raise (rb_eArgError, "[BUG] uninitialized surface for Ruby");
+      switch (cairo_surface_get_type (surface))
+        {
+        case CAIRO_SURFACE_TYPE_IMAGE:
+          klass = rb_cCairo_ImageSurface;
+          break;
+        case CAIRO_SURFACE_TYPE_PDF:
+          klass = rb_cCairo_PDFSurface;
+          break;
+        case CAIRO_SURFACE_TYPE_PS:
+          klass = rb_cCairo_PSSurface;
+          break;
+        case CAIRO_SURFACE_TYPE_SVG:
+          klass = rb_cCairo_SVGSurface;
+          break;
+        default:
+          klass = rb_cCairo_Surface;
+          break;
+        }
     }
+
   return klass;
 }
 
@@ -78,6 +99,7 @@ typedef struct cr_io_callback_closure {
   cairo_bool_t is_file;
 } cr_io_callback_closure_t;
 
+#if HAS_CREATE_CR_CLOSURE_SURFACE
 static VALUE
 cr_closure_target_push (VALUE klass, VALUE obj)
 {
@@ -147,6 +169,7 @@ cr_closure_free (void *closure)
 {
   cr_closure_destroy ((cr_io_callback_closure_t *) closure);
 }
+#endif
 
 static VALUE
 cr_surface_io_func_rescue (VALUE io_closure)
@@ -329,6 +352,19 @@ cr_surface_finish (VALUE self)
   return self;
 }
 
+static VALUE
+cr_surface_get_type (VALUE self)
+{
+  return INT2NUM (cairo_surface_get_type (_SELF));
+}
+
+static VALUE
+cr_surface_get_content (VALUE self)
+{
+  return INT2NUM (cairo_surface_get_content (_SELF));
+}
+
+
 #if CAIRO_HAS_PNG_FUNCTIONS
 static VALUE
 cr_image_surface_write_to_png_stream (VALUE self, VALUE target)
@@ -380,10 +416,9 @@ cr_image_surface_write_to_png_generic (VALUE self, VALUE target)
 static VALUE
 cr_surface_get_font_options (VALUE self)
 {
-  cairo_font_options_t *options = NULL;
+  cairo_font_options_t *options = cairo_font_options_create();
   cairo_surface_get_font_options (_SELF, options);
   cr_surface_check_status (_SELF);
-  options = cairo_font_options_copy (options);
   rb_cairo_check_status (cairo_font_options_status (options));
   return CRFONTOPTIONS2RVAL (options);
 }
@@ -434,6 +469,28 @@ cr_surface_set_device_offset (VALUE self, VALUE x_offset, VALUE y_offset)
   return self;
 }
 
+static VALUE
+cr_surface_get_device_offset (VALUE self)
+{
+  double x_offset, y_offset;
+
+  cairo_surface_get_device_offset (_SELF, &x_offset, &y_offset);
+  cr_surface_check_status (_SELF);
+  return rb_ary_new3 (2, rb_float_new (x_offset), rb_float_new (y_offset));
+}
+
+static VALUE
+cr_surface_set_fallback_resolution (VALUE self,
+                                    VALUE x_pixels_per_inch,
+                                    VALUE y_pixels_per_inch)
+{
+  cairo_surface_set_fallback_resolution (_SELF,
+                                         NUM2DBL (x_pixels_per_inch),
+                                         NUM2DBL (y_pixels_per_inch));
+  cr_surface_check_status (_SELF);
+  return self;
+}
+
 
 /* Image-surface functions */
 #if CAIRO_HAS_PNG_FUNCTIONS
@@ -532,6 +589,29 @@ cr_image_surface_initialize (int argc, VALUE *argv, VALUE self)
   return Qnil;
 }
 
+static VALUE
+cr_image_surface_get_data (VALUE self)
+{
+  unsigned char *data;
+  cairo_surface_t *surface;
+
+  surface = _SELF;
+  data = cairo_image_surface_get_data (surface);
+
+  if (data)
+    return rb_str_new ((const char *)data,
+                       cairo_image_surface_get_stride (surface) *
+                       cairo_image_surface_get_height (surface));
+  else
+    return Qnil;
+}
+
+static VALUE
+cr_image_surface_get_format (VALUE self)
+{
+  return INT2NUM (cairo_image_surface_get_format (_SELF));
+}
+
 static VALUE
 cr_image_surface_get_width (VALUE self)
 {
@@ -544,7 +624,14 @@ cr_image_surface_get_height (VALUE self)
   return INT2NUM (cairo_image_surface_get_height (_SELF));
 }
 
+static VALUE
+cr_image_surface_get_stride (VALUE self)
+{
+  return INT2NUM (cairo_image_surface_get_stride (_SELF));
+}
 
+
+/* Printing surfaces */
 #define DEFINE_SURFACE(type)                                              \
 static VALUE                                                              \
 cr_ ## type ## _surface_initialize (VALUE self, VALUE target,             \
@@ -585,28 +672,114 @@ cr_ ## type ## _surface_initialize (VALUE self, VALUE target,             \
   cr_surface_check_status (surface);                                      \
   DATA_PTR (self) = surface;                                              \
   return Qnil;                                                            \
-}                                                                         \
-                                                                          \
-static VALUE                                                              \
-cr_ ## type ## _surface_set_dpi (VALUE self, VALUE x_dpi, VALUE y_dpi)    \
-{                                                                         \
-  cairo_ ## type ## _surface_set_dpi (_SELF,                              \
-                                      NUM2DBL (x_dpi),                    \
-                                      NUM2DBL (y_dpi));                   \
-  cr_surface_check_status (_SELF);                                        \
-  return self;                                                            \
 }
 
+
 #if CAIRO_HAS_PS_SURFACE
 /* PS-surface functions */
 DEFINE_SURFACE(ps)
+
+static VALUE
+cr_ps_surface_set_size (VALUE self,
+                        VALUE width_in_points,
+                        VALUE height_in_points)
+{
+  cairo_ps_surface_set_size (_SELF,
+                             NUM2DBL (width_in_points),
+                             NUM2DBL (height_in_points));
+  cr_surface_check_status (_SELF);
+  return Qnil;
+}
+
+static VALUE
+cr_ps_surface_dsc_comment (VALUE self, VALUE comment)
+{
+  cairo_ps_surface_dsc_comment (_SELF, StringValueCStr (comment));
+  cr_surface_check_status (_SELF);
+  return Qnil;
+}
+
+static VALUE
+cr_ps_surface_dsc_begin_setup (VALUE self)
+{
+  cairo_ps_surface_dsc_begin_setup (_SELF);
+  cr_surface_check_status (_SELF);
+  if (rb_block_given_p ())
+    return rb_yield (self);
+  else
+    return Qnil;
+}
+
+static VALUE
+cr_ps_surface_dsc_begin_page_setup (VALUE self)
+{
+  cairo_ps_surface_dsc_begin_page_setup (_SELF);
+  cr_surface_check_status (_SELF);
+  if (rb_block_given_p ())
+    return rb_yield (self);
+  else
+    return Qnil;
+}
 #endif
 
 #if CAIRO_HAS_PDF_SURFACE
 /* PDF-surface functions */
 DEFINE_SURFACE(pdf)
+
+static VALUE
+cr_pdf_surface_set_size (VALUE self,
+                         VALUE width_in_points,
+                         VALUE height_in_points)
+{
+  cairo_pdf_surface_set_size (_SELF,
+                              NUM2DBL (width_in_points),
+                              NUM2DBL (height_in_points));
+  cr_surface_check_status (_SELF);
+  return Qnil;
+}
 #endif
 
+#if CAIRO_HAS_SVG_SURFACE
+/* SVG-surface functions */
+DEFINE_SURFACE(svg)
+
+static VALUE
+cr_svg_surface_restrict_to_version (VALUE self, VALUE version)
+{
+  cairo_svg_surface_restrict_to_version (_SELF, RVAL2CRSVGVERSION (version));
+  cr_surface_check_status (_SELF);
+  return Qnil;
+}
+
+static VALUE
+cr_svg_get_versions (VALUE self)
+{
+  VALUE rb_versions;
+  int i, num_versions;
+  cairo_svg_version_t const *versions;
+
+  cairo_svg_get_versions (&versions, &num_versions);
+
+  rb_versions = rb_ary_new2 (num_versions);
+
+  for (i = 0; i < num_versions; i++)
+    {
+      rb_ary_push (rb_versions, INT2NUM (versions[i]));
+    }
+
+  return rb_versions;
+}
+
+static VALUE
+cr_svg_version_to_string (VALUE self, VALUE version)
+{
+  const char *ver_str;
+  ver_str = cairo_svg_version_to_string (RVAL2CRSVGVERSION(version));
+  return rb_str_new2 (ver_str);
+}
+#endif
+
+
 void
 Init_cairo_surface (void)
 {
@@ -618,19 +791,24 @@ Init_cairo_surface (void)
   
   rb_cCairo_Surface =
     rb_define_class_under (rb_mCairo, "Surface", rb_cObject);
-
   rb_define_alloc_func (rb_cCairo_Surface, cr_surface_allocate);
   
   rb_define_method (rb_cCairo_Surface, "create_similar",
                     cr_surface_create_similar, 3);
   rb_define_method (rb_cCairo_Surface, "finish", cr_surface_finish, 0);
-  
+  rb_define_method (rb_cCairo_Surface, "type", cr_surface_get_type, 0);
+  rb_define_method (rb_cCairo_Surface, "content", cr_surface_get_content, 0);
+
   rb_define_method (rb_cCairo_Surface, "font_options",
                     cr_surface_get_font_options, 0);
   rb_define_method (rb_cCairo_Surface, "flush", cr_surface_flush, 0);
   rb_define_method (rb_cCairo_Surface, "mark_dirty", cr_surface_mark_dirty, 0);
   rb_define_method (rb_cCairo_Surface, "set_device_offset",
                     cr_surface_set_device_offset, 2);
+  rb_define_method (rb_cCairo_Surface, "device_offset",
+                    cr_surface_get_device_offset, 0);
+  rb_define_method (rb_cCairo_Surface, "set_fallback_resolution",
+                    cr_surface_set_fallback_resolution, 2);
 
   /* Image-surface */
   rb_cCairo_ImageSurface =
@@ -644,11 +822,17 @@ Init_cairo_surface (void)
   rb_define_method (rb_cCairo_ImageSurface, "initialize",
                     cr_image_surface_initialize, -1);
 
+  rb_define_method (rb_cCairo_ImageSurface, "data",
+                    cr_image_surface_get_data, 0);
+  rb_define_method (rb_cCairo_ImageSurface, "format",
+                    cr_image_surface_get_format, 0);
   rb_define_method (rb_cCairo_ImageSurface, "width",
                     cr_image_surface_get_width, 0);
   rb_define_method (rb_cCairo_ImageSurface, "height",
                     cr_image_surface_get_height, 0);
-  
+  rb_define_method (rb_cCairo_ImageSurface, "stride",
+                    cr_image_surface_get_stride, 0);
+
 #if CAIRO_HAS_PNG_FUNCTIONS
   rb_define_method (rb_cCairo_ImageSurface, "write_to_png",
                     cr_image_surface_write_to_png_generic, 1);
@@ -664,13 +848,19 @@ Init_cairo_surface (void)
                cr_id_holder, rb_hash_new ());                           \
                                                                         \
   rb_define_method (rb_cCairo_ ## name ## Surface, "initialize",        \
-                    cr_ ## type ## _surface_initialize, 3);             \
-  rb_define_method (rb_cCairo_ ## name ## Surface, "set_dpi",           \
-                    cr_ ## type ## _surface_set_dpi, 2);
+                    cr_ ## type ## _surface_initialize, 3);
 
 #if CAIRO_HAS_PS_SURFACE
   /* PS-surface */
   INIT_SURFACE(ps, PS)
+
+  rb_define_method (rb_cCairo_PSSurface, "set_size", cr_ps_surface_set_size, 2);
+  rb_define_method (rb_cCairo_PSSurface, "dsc_comment",
+                    cr_ps_surface_dsc_comment, 1);
+  rb_define_method (rb_cCairo_PSSurface, "dsc_begin_setup",
+                    cr_ps_surface_dsc_begin_setup, 0);
+  rb_define_method (rb_cCairo_PSSurface, "dsc_begin_page_setup",
+                    cr_ps_surface_dsc_begin_page_setup, 0);
 #else
   rb_cCairo_PSSurface = Qnil;
 #endif
@@ -678,7 +868,30 @@ Init_cairo_surface (void)
 #if CAIRO_HAS_PDF_SURFACE
   /* PDF-surface */
   INIT_SURFACE(pdf, PDF)
+
+  rb_define_method (rb_cCairo_PDFSurface, "set_size",
+                    cr_pdf_surface_set_size, 2);
 #else
   rb_cCairo_PDFSurface = Qnil;
 #endif
+
+#if CAIRO_HAS_SVG_SURFACE
+  /* SVG-surface */
+  INIT_SURFACE(svg, SVG)
+
+  rb_define_const (rb_cCairo_SVGSurface, "VERSION_1_1",
+                   rb_const_get (rb_mCairo, rb_intern ("SVG_VERSION_1_1")));
+  rb_define_const (rb_cCairo_SVGSurface, "VERSION_1_2",
+                   rb_const_get (rb_mCairo, rb_intern ("SVG_VERSION_1_2")));
+
+  rb_define_singleton_method (rb_cCairo_SVGSurface, "versions",
+                              cr_svg_get_versions, 0);
+  rb_define_singleton_method (rb_cCairo_SVGSurface, "version_to_string",
+                              cr_svg_version_to_string, 1);
+
+  rb_define_method (rb_cCairo_SVGSurface, "restrict_to_version",
+                    cr_svg_surface_restrict_to_version, 1);
+#else
+  rb_cCairo_SVGSurface = Qnil;
+#endif
 }
